#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""
Restate SDK for Python
"""

from .service import Service
from .object import VirtualObject
from .workflow import Workflow

# types
from .context import Context, ObjectContext, ObjectSharedContext
from .context import WorkflowContext, WorkflowSharedContext
from .context import DurablePromise

from .endpoint import app

try:
    from .harness import test_harness
except ImportError:
    # we don't have the appropriate dependencies installed

    # pylint: disable=unused-argument, redefined-outer-name
    def test_harness(app, follow_logs = False, restate_image = ""): # type: ignore
        """a dummy harness constructor to raise ImportError"""
        raise ImportError("Install restate-sdk[harness] to use this feature")



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""example.py"""
# pylint: disable=C0116
# pylint: disable=W0613

import os
from datetime import timedelta
from typing import (Dict, Iterable, List, Union, TypedDict, Literal, Any)
from restate import Service, Context

from . import list_object
from . import awakeable_holder

test_utils = Service("TestUtilsService")

@test_utils.handler()
async def echo(context: Context, input: str) -> str:
    return input

@test_utils.handler(name="uppercaseEcho")
async def uppercase_echo(context: Context, input: str) -> str:
    return input.upper()

@test_utils.handler(name="echoHeaders")
async def echo_headers(context: Context) -> Dict[str, str]:
    return context.request().headers

@test_utils.handler(name="createAwakeableAndAwaitIt")
async def create_awakeable_and_await_it(context: Context, req: Dict[str, Any]) -> Dict[str, Any]:
    name, awakeable = context.awakeable()

    await context.object_call(awakeable_holder.hold, key=req["awakeableKey"], arg=name)

    if "awaitTimeout" not in req:
        return {"type": "result", "value": await awakeable}

    timeout = context.sleep(timedelta(milliseconds=int(req["awaitTimeout"])))
    raise NotImplementedError()

@test_utils.handler(name="sleepConcurrently")
async def sleep_concurrently(context: Context, millis_duration: List[int]) -> None:
    timers = [context.sleep(timedelta(milliseconds=duration)) for duration in millis_duration]

    for timer in timers:
        await timer


@test_utils.handler(name="countExecutedSideEffects")
async def count_executed_side_effects(context: Context, increments: int) -> int:
    invoked_side_effects = 0

    def effect():
        nonlocal invoked_side_effects
        invoked_side_effects += 1

    for _ in range(increments):
        await context.run("count", effect)

    return invoked_side_effects

@test_utils.handler(name="getEnvVariable")
async def get_env_variable(context: Context, env_name: str) -> str:
    return os.environ.get(env_name, default="")

class CreateAwakeableAndAwaitIt(TypedDict):
    type: Literal["createAwakeableAndAwaitIt"]
    awakeableKey: str

class GetEnvVariable(TypedDict):
    type: Literal["getEnvVariable"]
    envName: str

Command = Union[
    CreateAwakeableAndAwaitIt,
    GetEnvVariable
]

class InterpretRequest(TypedDict):
    listName: str
    commands: Iterable[Command]

@test_utils.handler(name="interpretCommands")
async def interpret_commands(context: Context, req: InterpretRequest):
    for cmd in req['commands']:
        if cmd['type'] == "createAwakeableAndAwaitIt":
            name, awakeable = context.awakeable()
            context.object_send(awakeable_holder.hold, key=cmd["awakeableKey"], arg=name)
            result = await awakeable
            context.object_send(list_object.append, key=req['listName'], arg=result)
        elif cmd['type'] == "getEnvVariable":
            context.object_send(list_object.append, key=req['listName'], arg=os.environ.get(cmd['envName'], default=""))



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#

# pylint: disable=R0917
"""
This module contains the definition of the Handler class,
which is used to define the handlers for the services.
"""

from dataclasses import dataclass
from inspect import Signature
from typing import Any, Callable, Awaitable, Generic, Literal, Optional, TypeVar

from restate.exceptions import TerminalError
from restate.serde import JsonSerde, Serde, PydanticJsonSerde

I = TypeVar('I')
O = TypeVar('O')
T = TypeVar('T')

# we will use this symbol to store the handler in the function
RESTATE_UNIQUE_HANDLER_SYMBOL = str(object())


def try_import_pydantic_base_model():
    """
    Try to import PydanticBaseModel from Pydantic.
    """
    try:
        from pydantic import BaseModel # type: ignore # pylint: disable=import-outside-toplevel
        return BaseModel
    except ImportError:
        class Dummy: # pylint: disable=too-few-public-methods
            """a dummy class to use when Pydantic is not available"""

        return Dummy

PydanticBaseModel = try_import_pydantic_base_model()

@dataclass
class ServiceTag:
    """
    This class is used to identify the service.
    """
    kind: Literal["object", "service", "workflow"]
    name: str

@dataclass
class TypeHint(Generic[T]):
    """
    Represents a type hint.
    """
    annotation: Optional[T] = None
    is_pydantic: bool = False

@dataclass
class HandlerIO(Generic[I, O]):
    """
    Represents the input/output configuration for a handler.

    Attributes:
        accept (str): The accept header value for the handler.
        content_type (str): The content type header value for the handler.
    """
    accept: str
    content_type: str
    input_serde: Serde[I]
    output_serde: Serde[O]
    input_type: Optional[TypeHint[I]] = None
    output_type: Optional[TypeHint[O]] = None

def is_pydantic(annotation) -> bool:
    """
    Check if an object is a Pydantic model.
    """
    try:
        return issubclass(annotation, PydanticBaseModel)
    except TypeError:
        # annotation is not a class or a type
        return False


def extract_io_type_hints(handler_io: HandlerIO[I, O], signature: Signature):
    """
    Augment handler_io with additional information about the input and output types.

    This function has a special check for Pydantic models when these are provided.
    This method will inspect the signature of an handler and will look for
    the input and the return types of a function, and will:
    * capture any Pydantic models (to be used later at discovery)
    * replace the default json serializer (is unchanged by a user) with a Pydantic serde
    """
    annotation = list(signature.parameters.values())[-1].annotation
    handler_io.input_type = TypeHint(annotation=annotation, is_pydantic=False)

    if is_pydantic(annotation):
        handler_io.input_type.is_pydantic = True
        if isinstance(handler_io.input_serde, JsonSerde): # type: ignore
            handler_io.input_serde = PydanticJsonSerde(annotation)

    annotation = signature.return_annotation
    handler_io.output_type = TypeHint(annotation=annotation, is_pydantic=False)

    if is_pydantic(annotation):
        handler_io.output_type.is_pydantic=True
        if isinstance(handler_io.output_serde, JsonSerde): # type: ignore
            handler_io.output_serde = PydanticJsonSerde(annotation)

@dataclass
class Handler(Generic[I, O]):
    """
    Represents a handler for a service.
    """
    service_tag: ServiceTag
    handler_io: HandlerIO[I, O]
    kind: Optional[Literal["exclusive", "shared", "workflow"]]
    name: str
    fn: Callable[[Any, I], Awaitable[O]] | Callable[[Any], Awaitable[O]]
    arity: int


# disable too many arguments warning
# pylint: disable=R0913

def make_handler(service_tag: ServiceTag,
                 handler_io: HandlerIO[I, O],
                 name: str | None,
                 kind: Optional[Literal["exclusive", "shared", "workflow"]],
                 wrapped: Any,
                 signature: Signature) -> Handler[I, O]:
    """
    Factory function to create a handler.
    """
    # try to deduce the handler name
    handler_name = name
    if not handler_name:
        handler_name = wrapped.__name__
    if not handler_name:
        raise ValueError("Handler name must be provided")

    if len(signature.parameters) == 0:
        raise ValueError("Handler must have at least one parameter")

    arity = len(signature.parameters)
    extract_io_type_hints(handler_io, signature)

    handler = Handler[I, O](service_tag,
                            handler_io,
                            kind,
                            handler_name,
                            wrapped,
                            arity)

    vars(wrapped)[RESTATE_UNIQUE_HANDLER_SYMBOL] = handler
    return handler

def handler_from_callable(wrapper: Callable[[Any, I], Awaitable[O]]) -> Handler[I, O]:
    """
    Get the handler from the callable.
    """
    try:
        return vars(wrapper)[RESTATE_UNIQUE_HANDLER_SYMBOL]
    except KeyError:
        raise ValueError("Handler not found") # pylint: disable=raise-missing-from

async def invoke_handler(handler: Handler[I, O], ctx: Any, in_buffer: bytes) -> bytes:
    """
    Invoke the handler with the given context and input.
    """
    if handler.arity == 2:
        try:
            in_arg = handler.handler_io.input_serde.deserialize(in_buffer) # type: ignore
        except Exception as e:
            raise TerminalError(message=f"Unable to parse an input argument. {e}") from e
        out_arg = await handler.fn(ctx, in_arg) # type: ignore [call-arg, arg-type]
    else:
        out_arg = await handler.fn(ctx) # type: ignore [call-arg]
    out_buffer = handler.handler_io.output_serde.serialize(out_arg) # type: ignore
    return bytes(out_buffer)



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""greeter.py"""
# pylint: disable=C0116
# pylint: disable=W0613
# pylint: disable=C0115
# pylint: disable=R0903

from pydantic import BaseModel
from restate import Service, Context

# models
class GreetingRequest(BaseModel):
    name: str

class Greeting(BaseModel):
    message: str

# service

pydantic_greeter = Service("pydantic_greeter")

@pydantic_greeter.handler()
async def greet(ctx: Context, req: GreetingRequest) -> Greeting:
    return Greeting(message=f"Hello {req.name}!")



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
""" This module contains functions for serializing and deserializing data. """
import abc
import json
import typing

T = typing.TypeVar('T')
I = typing.TypeVar('I')
O = typing.TypeVar('O')

# disable to few parameters
# pylint: disable=R0903

class SerializerType(typing.Generic[O]):
    """A type definition for a serializer"""
    __call__: typing.Callable[[typing.Optional[O]], bytes]

class DeserializerType(typing.Generic[I]):
    """A type definition for a deserializer"""
    __call__: typing.Callable[[bytes], typing.Optional[I]]


class Serde(typing.Generic[T], abc.ABC):
    """serializer/deserializer interface."""

    @abc.abstractmethod
    def deserialize(self, buf: bytes) -> typing.Optional[T]:
        """
        Deserializes a bytearray to an object.
        """

    @abc.abstractmethod
    def serialize(self, obj: typing.Optional[T]) -> bytes:
        """
        Serializes an object to a bytearray.
        """


class BytesSerde(Serde[bytes]):
    """A pass-trough serializer/deserializer."""

    def deserialize(self, buf: bytes) -> typing.Optional[bytes]:
        """
        Deserializes a bytearray to a bytearray.

        Args:
            buf (bytearray): The bytearray to deserialize.

        Returns:
            typing.Optional[bytes]: The deserialized bytearray.
        """
        return buf

    def serialize(self, obj: typing.Optional[bytes]) -> bytes:
        """
        Serializes a bytearray to a bytearray.

        Args:
            obj (bytes): The bytearray to serialize.

        Returns:
            bytearray: The serialized bytearray.
        """
        if obj is None:
            return bytes()
        return obj


class JsonSerde(Serde[I]):
    """A JSON serializer/deserializer."""

    def deserialize(self, buf: bytes) -> typing.Optional[I]:
        """
        Deserializes a bytearray to a JSON object.

        Args:
            buf (bytearray): The bytearray to deserialize.

        Returns:
            typing.Optional[I]: The deserialized JSON object.
        """
        if not buf:
            return None
        return json.loads(buf)

    def serialize(self, obj: typing.Optional[I]) -> bytes:
        """
        Serializes a JSON object to a bytearray.

        Args:
            obj (I): The JSON object to serialize.

        Returns:
            bytearray: The serialized bytearray.
        """
        if obj is None:
            return bytes()

        return bytes(json.dumps(obj), "utf-8")


class PydanticJsonSerde(Serde[I]):
    """
    Serde for Pydantic models to/from JSON
    """

    def __init__(self, model):
        self.model = model

    def deserialize(self, buf: bytes) -> typing.Optional[I]:
        """
        Deserializes a bytearray to a Pydantic model.

        Args:
            buf (bytearray): The bytearray to deserialize.

        Returns:
            typing.Optional[I]: The deserialized Pydantic model.
        """
        if not buf:
            return None
        return self.model.model_validate_json(buf)

    def serialize(self, obj: typing.Optional[I]) -> bytes:
        """
        Serializes a Pydantic model to a bytearray.

        Args:
            obj (I): The Pydantic model to serialize.

        Returns:
            bytearray: The serialized bytearray.
        """
        if obj is None:
            return bytes()
        json_str = obj.model_dump_json() # type: ignore[attr-defined]
        return json_str.encode("utf-8")

def deserialize_json(buf: typing.ByteString) -> typing.Optional[O]:
    """
    Deserializes a bytearray to a JSON object.

    Args:
        buf (bytearray): The bytearray to deserialize.

    Returns:
        typing.Optional[O]: The deserialized JSON object.
    """
    if not buf:
        return None
    return json.loads(buf)

def serialize_json(obj: typing.Optional[O]) -> bytes:
    """
    Serializes a JSON object to a bytearray.

    Args:
        obj (O): The JSON object to serialize.

    Returns:
        bytearray: The serialized bytearray.
    """
    if obj is None:
        return bytes()

    return bytes(json.dumps(obj), "utf-8")



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""example.py"""
# pylint: disable=C0116
# pylint: disable=W0613

from datetime import timedelta
from typing import Dict
from restate import VirtualObject, ObjectContext

from . import counter

invoke_counts: Dict[str, int] = {}

def do_left_action(ctx: ObjectContext) -> bool:
    count_key = ctx.key()
    invoke_counts[count_key] = invoke_counts.get(count_key, 0) + 1
    return invoke_counts[count_key] % 2 == 1

def increment_counter(ctx: ObjectContext):
    ctx.object_send(counter.add, key=ctx.key(), arg=1)

non_deterministic = VirtualObject("NonDeterministic")

@non_deterministic.handler(name="setDifferentKey")
async def set_different_key(ctx: ObjectContext):
    if do_left_action(ctx):
        ctx.set("a", "my-state")
    else:
        ctx.set("b", "my-state")
    await ctx.sleep(timedelta(milliseconds=100))
    increment_counter(ctx)

@non_deterministic.handler(name="backgroundInvokeWithDifferentTargets")
async def background_invoke_with_different_targets(ctx: ObjectContext):
    if do_left_action(ctx):
        ctx.object_send(counter.get, key="abc", arg=None)
    else:
        ctx.object_send(counter.reset, key="abc", arg=None)
    await ctx.sleep(timedelta(milliseconds=100))
    increment_counter(ctx)

@non_deterministic.handler(name="callDifferentMethod")
async def call_different_method(ctx: ObjectContext):
    if do_left_action(ctx):
        await ctx.object_call(counter.get, key="abc", arg=None)
    else:
        await ctx.object_call(counter.reset, key="abc", arg=None)
    await ctx.sleep(timedelta(milliseconds=100))
    increment_counter(ctx)

@non_deterministic.handler(name="eitherSleepOrCall")
async def either_sleep_or_call(ctx: ObjectContext):
    if do_left_action(ctx):
        await ctx.sleep(timedelta(milliseconds=100))
    else:
        await ctx.object_call(counter.get, key="abc", arg=None)
    await ctx.sleep(timedelta(milliseconds=100))
    increment_counter(ctx)



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""This module contains the restate exceptions"""

class TerminalError(Exception):
    """This exception is raised to indicate a termination of the execution"""

    def __init__(self, message: str, status_code: int = 500) -> None:
        super().__init__(message)
        self.message = message
        self.status_code = status_code



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""
wrap the restate._internal.PyVM class
"""
# pylint: disable=E1101,R0917

from dataclasses import dataclass
import typing
from restate._internal import PyVM, PyFailure, PySuspended, PyVoid, PyStateKeys, PyExponentialRetryConfig  # pylint: disable=import-error,no-name-in-module,line-too-long

@dataclass
class Invocation:
    """
    Invocation dataclass
    """
    invocation_id: str
    random_seed: int
    headers: typing.List[typing.Tuple[str, str]]
    input_buffer: bytes
    key: str

@dataclass
class RunRetryConfig:
    """
    Expo Retry Configuration
    """
    initial_interval: typing.Optional[int] = None
    max_attempts: typing.Optional[int] = None
    max_duration: typing.Optional[int] = None

@dataclass
class Failure:
    """
    Failure
    """
    code: int
    message: str

@dataclass
class NotReady:
    """
    NotReady
    """

class SuspendedException(Exception):
    """
    Suspended Exception
    """
    def __init__(self, *args: object) -> None:
        super().__init__(*args)

NOT_READY = NotReady()
SUSPENDED = SuspendedException()

AsyncResultType = typing.Optional[typing.Union[bytes, Failure, NotReady]]

# pylint: disable=too-many-public-methods
class VMWrapper:
    """
    A wrapper class for the restate_sdk._internal.PyVM class.
    It provides a type-friendly interface to our shared vm.
    """

    def __init__(self, headers: typing.List[typing.Tuple[str, str]]):
        self.vm = PyVM(headers)

    def get_response_head(self) -> typing.Tuple[int, typing.Iterable[typing.Tuple[str, str]]]:
        """
        Retrieves the response head from the virtual machine.

        Returns:
            A tuple containing the status code and a list of header tuples.
        """
        result = self.vm.get_response_head()
        return (result.status_code, result.headers)

    def notify_input(self, input_buf: bytes):
        """Send input to the virtual machine."""
        self.vm.notify_input(input_buf)

    def notify_input_closed(self):
        """Notify the virtual machine that the input has been closed."""
        self.vm.notify_input_closed()

    def notify_error(self, error: str):
        """Notify the virtual machine of an error."""
        self.vm.notify_error(error)

    def take_output(self) -> typing.Optional[bytes]:
        """Take the output from the virtual machine."""
        return self.vm.take_output()

    def notify_await_point(self, handle: int):
        """Notify the virtual machine of an await point."""
        self.vm.notify_await_point(handle)

    def is_ready_to_execute(self) -> bool:
        """Returns true when the VM is ready to operate."""
        return self.vm.is_ready_to_execute()

    def take_async_result(self, handle: typing.Any) -> AsyncResultType:
        """Take the result of an asynchronous operation."""
        result = self.vm.take_async_result(handle)
        if result is None:
            return NOT_READY
        if isinstance(result, PyVoid):
            # success with an empty value
            return None
        if isinstance(result, bytes):
            # success with a non empty value
            return result
        if isinstance(result, PyStateKeys):
            # success with state keys
            return result.keys
        if isinstance(result, PyFailure):
            # a terminal failure
            code = result.code
            message = result.message
            return Failure(code, message)
        if isinstance(result, PySuspended):
            # the state machine had suspended
            raise SUSPENDED
        raise ValueError(f"Unknown result type: {result}")

    def sys_input(self) -> Invocation:
        """
            Retrieves the system input from the virtual machine.

            Returns:
                An instance of the Invocation class containing the system input.
        """
        inp = self.vm.sys_input()
        invocation_id: str = inp.invocation_id
        random_seed: int = inp.random_seed
        headers: typing.List[typing.Tuple[str, str]] = [(h.key, h.value) for h in inp.headers]
        input_buffer: bytes = bytes(inp.input)
        key: str = inp.key

        return Invocation(
            invocation_id=invocation_id,
            random_seed=random_seed,
            headers=headers,
            input_buffer=input_buffer,
            key=key)

    def sys_write_output_success(self, output: bytes):
        """
        Writes the output to the system.

        Args:
          output: The output to be written. It can be either a bytes or a Failure object.

        Returns:
            None
        """
        self.vm.sys_write_output_success(output)

    def sys_write_output_failure(self, output: Failure):
        """
        Writes the output to the system.

        Args:
          output: The output to be written. It can be either a bytes or a Failure object.

        Returns:
            None
        """
        res = PyFailure(output.code, output.message)
        self.vm.sys_write_output_failure(res)


    def sys_get_state(self, name) -> int:
        """
        Retrieves a key-value binding.

        Args:
            name: The name of the value to be retrieved.

        Returns:
            The value associated with the given name.
        """
        return self.vm.sys_get_state(name)


    def sys_get_state_keys(self) -> int:
        """
        Retrieves all keys.

        Returns:
            The state keys
        """
        return self.vm.sys_get_state_keys()


    def sys_set_state(self, name: str, value: bytes):
        """
        Sets a key-value binding.

        Args:
            name: The name of the value to be set.
            value: The value to be set.

        Returns:
            None
        """
        self.vm.sys_set_state(name, value)

    def sys_clear_state(self, name: str):
        """Clear the state associated with the given name."""
        self.vm.sys_clear_state(name)

    def sys_clear_all_state(self):
        """Clear the state associated with the given name."""
        self.vm.sys_clear_all_state()

    def sys_sleep(self, millis: int):
        """Ask to sleep for a given duration"""
        return self.vm.sys_sleep(millis)

    def sys_call(self,
                 service: str,
                 handler: str,
                 parameter: bytes,
                 key: typing.Optional[str] = None):
        """Call a service"""
        return self.vm.sys_call(service, handler, parameter, key)

    # pylint: disable=too-many-arguments
    def sys_send(self,
                 service: str,
                 handler: str,
                 parameter: bytes,
                 key: typing.Optional[str] = None,
                 delay: typing.Optional[int] = None) -> None:
        """send an invocation to a service (no response)"""
        self.vm.sys_send(service, handler, parameter, key, delay)

    def sys_run_enter(self, name: str) -> typing.Union[bytes, None, Failure]:
        """
        Enter a side effect

        Returns:
            None if the side effect was not journald.
            PyFailure if the side effect failed.
            bytes if the side effect was successful.
        """
        result = self.vm.sys_run_enter(name)
        if result is None:
            return None
        if isinstance(result, PyFailure):
            return Failure(result.code, result.message) # pylint: disable=protected-access
        assert isinstance(result, bytes)
        return result

    def sys_awakeable(self) -> typing.Tuple[str, int]:
        """
        Return a fresh awaitable
        """
        return self.vm.sys_awakeable()

    def sys_resolve_awakeable(self, name: str, value: bytes):
        """
        Resolve
        """
        self.vm.sys_complete_awakeable_success(name, value)

    def sys_reject_awakeable(self, name: str, failure: Failure):
        """
        Reject
        """
        py_failure = PyFailure(failure.code, failure.message)
        self.vm.sys_complete_awakeable_failure(name, py_failure)

    def sys_run_exit_success(self, output: bytes) -> int:
        """
        Exit a side effect

        Args:
            output: The output of the side effect.

        Returns:
            handle
        """
        return self.vm.sys_run_exit_success(output)

    def sys_get_promise(self, name: str) -> int:
        """Returns the promise handle"""
        return self.vm.sys_get_promise(name)

    def sys_peek_promise(self, name: str) -> int:
        """Peek into the workflow promise"""
        return self.vm.sys_peek_promise(name)

    def sys_complete_promise_success(self, name: str, value: bytes) -> int:
        """Complete the promise"""
        return self.vm.sys_complete_promise_success(name, value)

    def sys_complete_promise_failure(self, name: str, failure: Failure) -> int:
        """reject the promise on failure"""
        res = PyFailure(failure.code, failure.message)
        return self.vm.sys_complete_promise_failure(name, res)

    def sys_run_exit_failure(self, output: Failure) -> int:
        """
        Exit a side effect

        Args:
            name: The name of the side effect.
            output: The output of the side effect.
        """
        res = PyFailure(output.code, output.message)
        return self.vm.sys_run_exit_failure(res)

    # pylint: disable=line-too-long
    def sys_run_exit_transient(self, failure: Failure, attempt_duration_ms: int, config: RunRetryConfig) -> int | None:
        """
        Exit a side effect with a transient Error.
        This requires a retry policy to be provided.
        """
        py_failure = PyFailure(failure.code, failure.message)
        py_config = PyExponentialRetryConfig(config.initial_interval, config.max_attempts, config.max_duration)
        try:
            handle = self.vm.sys_run_exit_failure_transient(py_failure, attempt_duration_ms, py_config)
            # The VM decided not to retry, therefore we get back an handle that will be resolved
            # with a terminal failure.
            return handle
        # pylint: disable=bare-except
        except:
            # The VM decided to retry, therefore we tear down the current execution
            return None

    def sys_end(self):
        """
        This method is responsible for ending the system.

        It calls the `sys_end` method of the `vm` object.
        """
        self.vm.sys_end()



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""Test containers based wrapper for the restate server"""

import asyncio
from dataclasses import dataclass
import threading
import typing
from urllib.error import URLError
import socket

from hypercorn.config import Config
from hypercorn.asyncio import serve
from testcontainers.core.container import DockerContainer # type: ignore
from testcontainers.core.waiting_utils import wait_container_is_ready # type: ignore
import httpx


def find_free_port():
    """find the next free port to bind to on the host machine"""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("0.0.0.0", 0))
        return s.getsockname()[1]

def run_in_background(coro) -> threading.Thread:
    """run a coroutine in the background"""
    def runner():
        asyncio.run(coro)

    thread = threading.Thread(target=runner, daemon=True)
    thread.start()
    return thread


class BindAddress:
    """A bind address for the ASGI server"""

    def get_local_bind_address(self):
        """return the local bind address for hypercorn to bind to"""

    def get_endpoint_connection_string(self):
        """return the SDK connection string to be used by restate"""

    def cleanup(self):
        """cleanup any resources used by the bind address"""

class TcpSocketBindAddress(BindAddress):
    """Bind a TCP address that listens on a random TCP port"""

    def __init__(self):
        self.port = find_free_port()

    def get_local_bind_address(self):
        return f"0.0.0.0:{self.port}"

    def get_endpoint_connection_string(self):
        return f"http://host.docker.internal:{self.port}"


class AsgiServer:
    """A simple ASGI server that listens on a unix domain socket"""

    thread: typing.Optional[threading.Thread] = None

    def __init__(self, asgi_app, bind_address: BindAddress):
        self.asgi_app = asgi_app
        self.bind_address = bind_address
        self.stop_event = asyncio.Event()
        self.exit_event = asyncio.Event()

    def stop(self):
        """stop the server"""
        self.stop_event.set()
        if self.thread:
            self.thread.join(timeout=1)
            self.thread = None
        self.exit_event.set()

    def start(self):
        """start the server"""

        def shutdown_trigger():
            """trigger the shutdown event"""
            return self.stop_event.wait()

        async def run_asgi():
            """run the asgi app on the given port"""
            config = Config()
            config.h2_max_concurrent_streams = 2147483647
            config.keep_alive_max_requests = 2147483647
            config.keep_alive_timeout = 2147483647

            bind = self.bind_address.get_local_bind_address()
            config.bind = [bind]
            try:
                print(f"Starting ASGI server on {bind}", flush=True)
                await serve(self.asgi_app,
                            config=config,
                            mode='asgi',
                            shutdown_trigger=shutdown_trigger)
            except asyncio.CancelledError:
                print("ASGI server was cancelled", flush=True)
            except Exception as e: # pylint: disable=broad-except
                print(f"Failed to start the ASGI server: {e}", flush=True)
                raise e
            finally:
                self.exit_event.set()

        self.thread = run_in_background(run_asgi())
        return self

class RestateContainer(DockerContainer):
    """Create a Restate container"""

    log_thread: typing.Optional[threading.Thread] = None

    def __init__(self, image):
        super().__init__(image)
        self.with_exposed_ports(8080, 9070)
        self.with_env('RESTATE_LOG_FILTER', 'restate=info')
        self.with_env('RESTATE_BOOTSTRAP_NUM_PARTITIONS', '1')
        self.with_env('RESTATE_DEFAULT_NUM_PARTITIONS', '1')
        self.with_env('RESTATE_SHUTDOWN_TIMEOUT', '10s')
        self.with_env('RESTATE_ROCKSDB_TOTAL_MEMORY_SIZE', '32 MB')
        self.with_env('RESTATE_WORKER__INVOKER__IN_MEMORY_QUEUE_LENGTH_LIMIT', '64')
        self.with_env('RESTATE_WORKER__INVOKER__INACTIVITY_TIMEOUT', '10m')
        self.with_env('RESTATE_WORKER__INVOKER__ABORT_TIMEOUT', '10m')

        self.with_kwargs(extra_hosts={"host.docker.internal" : "host-gateway"})

    def ingress_url(self):
        """return the URL to access the Restate ingress"""
        return f"http://{self.get_container_host_ip()}:{self.get_exposed_port(8080)}"

    def admin_url(self):
        """return the URL to access the Restate admin"""
        return f"http://{self.get_container_host_ip()}:{self.get_exposed_port(9070)}"

    def get_admin_client(self):
        """return an httpx client to access the admin interface"""
        return httpx.Client(base_url=self.admin_url())

    def get_ingress_client(self):
        """return an httpx client to access the ingress interface"""
        return httpx.Client(base_url=self.ingress_url())

    @wait_container_is_ready(httpx.HTTPError, URLError)
    def _wait_healthy(self):
        """wait for restate's health checks to pass"""
        self.get_ingress_client().get("/restate/health").raise_for_status()
        self.get_admin_client().get("/health").raise_for_status()


    def start(self, stream_logs = False):
        """start the container and wait for health checks to pass"""
        super().start()

        def stream_log():
            for line in self.get_wrapped_container().logs(stream=True):
                print(line.decode("utf-8"), end="", flush=True)

        if stream_logs:
            thread = threading.Thread(target=stream_log, daemon=True)
            thread.start()
            self.log_thread = thread

        self._wait_healthy()
        return self


@dataclass
class TestConfiguration:
    """A configuration for running tests"""
    restate_image: str = "restatedev/restate:latest"
    stream_logs: bool = False


class RestateTestHarness:
    """A test harness for running Restate SDKs"""
    bind_address: typing.Optional[BindAddress] = None
    server: typing.Optional[AsgiServer] = None
    restate: typing.Optional[RestateContainer] = None

    def __init__(self, asgi_app, config: typing.Optional[TestConfiguration]):
        self.asgi_app = asgi_app
        if config:
            self.config = config
        else:
            self.config = TestConfiguration()

    def start(self):
        """start the restate server and the sdk"""
        self.bind_address = TcpSocketBindAddress()
        self.server = AsgiServer(self.asgi_app, self.bind_address).start()
        self.restate = RestateContainer(image=self.config.restate_image) \
            .start(self.config.stream_logs)
        try:
            self._register_sdk()
        except Exception as e:
            self.stop()
            raise AssertionError("Failed to register the SDK with the Restate server") from e

    def _register_sdk(self):
        """register the sdk with the restate server"""
        assert self.bind_address is not None
        assert self.restate is not None

        uri = self.bind_address.get_endpoint_connection_string()
        client = self.restate.get_admin_client()
        res = client.post("/deployments",
                          headers={"content-type" : "application/json"},
                          json={"uri": uri})
        if not res.is_success:
            msg = f"unable to register the services at {uri} - {res.status_code} {res.text}"
            raise AssertionError(msg)

    def stop(self):
        """terminate the restate server and the sdk"""
        if self.server is not None:
            self.server.stop()
        if self.restate is not None:
            self.restate.stop()
        if self.bind_address is not None:
            self.bind_address.cleanup()

    def ingress_client(self):
        """return an httpx client to access the restate server's ingress"""
        if self.restate is None:
            raise AssertionError("The Restate server has not been started. Use .start()")
        return self.restate.get_ingress_client()


    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()
        return False


def test_harness(app,
                 follow_logs: bool = False,
                 restate_image: str = "restatedev/restate:latest") -> RestateTestHarness:
    """create a test harness for running Restate SDKs"""
    config = TestConfiguration(
        restate_image=restate_image,
        stream_logs=follow_logs,
    )
    return RestateTestHarness(app, config)



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""example.py"""
# pylint: disable=C0116
# pylint: disable=W0613

from restate import Service, Context, VirtualObject, ObjectContext

from . import awakeable_holder

kill_runner = Service("KillTestRunner")

@kill_runner.handler(name="startCallTree")
async def start_call_tree(ctx: Context):
    await ctx.object_call(recursive_call, key="", arg=None)

kill_singleton = VirtualObject("KillTestSingleton")

@kill_singleton.handler(name="recursiveCall")
async def recursive_call(ctx: ObjectContext):
    name, promise = ctx.awakeable()
    ctx.object_send(awakeable_holder.hold, key="kill", arg=name)
    await promise

    await ctx.object_call(recursive_call, key="", arg=None)

@kill_singleton.handler(name="isUnlocked")
async def is_unlocked(ctx: ObjectContext):
    return None



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""This module contains the ASGI server for the restate framework."""

import asyncio
from typing import Dict, Literal
import traceback
from restate.discovery import compute_discovery_json
from restate.endpoint import Endpoint
from restate.server_context import ServerInvocationContext
from restate.server_types import Receive, Scope, Send, binary_to_header, header_to_binary
from restate.vm import VMWrapper
from restate._internal import PyIdentityVerifier, IdentityVerificationException # pylint: disable=import-error,no-name-in-module
from restate._internal import SDK_VERSION # pylint: disable=import-error,no-name-in-module
from restate.aws_lambda import is_running_on_lambda, wrap_asgi_as_lambda_handler

X_RESTATE_SERVER = header_to_binary([("x-restate-server", f"restate-sdk-python/{SDK_VERSION}")])

async def send_status(send, receive, status_code: int):
    """respond with a status code"""
    await send({'type': 'http.response.start', 'status': status_code, "headers": X_RESTATE_SERVER})
    # For more info on why this loop, see ServerInvocationContext.leave()
    # pylint: disable=R0801
    while True:
        event = await receive()
        if event is None:
            break
        if event.get('type') == 'http.disconnect':
            break
        if event.get('type') == 'http.request' and event.get('more_body', False) is False:
            break
    await send({'type': 'http.response.body'})

async def send404(send, receive):
    """respond with a 404"""
    await send_status(send, receive, 404)

async def send_discovery(scope: Scope, send: Send, endpoint: Endpoint):
    """respond with a discovery"""
    discovered_as: Literal["request_response", "bidi"]
    if scope['http_version'] == '1.1':
        discovered_as = "request_response"
    else:
        discovered_as = "bidi"
    headers, js = compute_discovery_json(endpoint, 1, discovered_as)
    bin_headers = header_to_binary(headers.items())
    bin_headers.extend(X_RESTATE_SERVER)
    await send({
        'type': 'http.response.start',
        'status': 200,
        'headers': bin_headers,
        'trailers': False
        })
    await send({
        'type': 'http.response.body',
        'body': js.encode('utf-8'),
        'more_body': False,
    })

async def process_invocation_to_completion(vm: VMWrapper,
                                           handler,
                                           attempt_headers: Dict[str, str],
                                           receive: Receive,
                                           send: Send):
    """Invoke the user code."""
    status, res_headers = vm.get_response_head()
    res_bin_headers = header_to_binary(res_headers)
    res_bin_headers.extend(X_RESTATE_SERVER)
    await send({
        'type': 'http.response.start',
        'status': status,
        'headers': res_bin_headers,
        'trailers': False
    })
    assert status == 200
    # ========================================
    # Read the input and the journal
    # ========================================
    while True:
        message = await receive()
        if message.get('type') == 'http.disconnect':
            # everything ends here really ...
            return
        if message.get('type') == 'http.request':
            assert isinstance(message['body'], bytes)
            vm.notify_input(message['body'])
        if not message.get('more_body', False):
            vm.notify_input_closed()
            break
        if vm.is_ready_to_execute():
            break
    # ========================================
    # Execute the user code
    # ========================================
    invocation = vm.sys_input()
    context = ServerInvocationContext(vm=vm,
                                      handler=handler,
                                      invocation=invocation,
                                      attempt_headers=attempt_headers,
                                      send=send,
                                      receive=receive)
    try:
        await context.enter()
    except asyncio.exceptions.CancelledError:
        raise
    # pylint: disable=W0718
    except Exception:
        traceback.print_exc()
    await context.leave()

class LifeSpanNotImplemented(ValueError):
    """Signal to the asgi server that we didn't implement lifespans"""

def asgi_app(endpoint: Endpoint):
    """Create an ASGI-3 app for the given endpoint."""

    # Prepare request signer
    identity_verifier = PyIdentityVerifier(endpoint.identity_keys)

    async def app(scope: Scope, receive: Receive, send: Send):
        try:
            if scope['type'] == 'lifespan':
                raise LifeSpanNotImplemented()
            if scope['type'] != 'http':
                raise NotImplementedError(f"Unknown scope type {scope['type']}")

            request_path = scope['path']
            assert isinstance(request_path, str)

            # Verify Identity
            assert not isinstance(scope['headers'], str)
            assert hasattr(scope['headers'], '__iter__')
            request_headers = binary_to_header(scope['headers'])
            try:
                identity_verifier.verify(request_headers, request_path)
            except IdentityVerificationException:
                # Identify verification failed, send back unauthorized and close
                await send_status(send, receive,401)
                return

            # might be a discovery request
            if request_path == '/discover':
                await send_discovery(scope, send, endpoint)
                return
            # anything other than invoke is 404
            if not request_path.startswith('/invoke/'):
                await send404(send, receive)
                return
            # path is of the form: /invoke/:service/:handler
            # strip "/invoke/" (= strlen 8) and split the service and handler
            service_name, handler_name = request_path[8:].split('/')
            service = endpoint.services.get(service_name)
            if not service:
                await send404(send, receive)
                return
            handler = service.handlers.get(handler_name)
            if not handler:
                await send404(send, receive)
                return
            #
            # At this point we have a valid handler.
            # Let us setup restate's execution context for this invocation and handler.
            #
            await process_invocation_to_completion(VMWrapper(request_headers),
                                                   handler,
                                                   dict(request_headers),
                                                   receive,
                                                   send)
        except LifeSpanNotImplemented as e:
            raise e
        except Exception as e:
            traceback.print_exc()
            raise e

    if is_running_on_lambda():
        # If we're on Lambda, just return the adapter
        return wrap_asgi_as_lambda_handler(app)

    return app



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""testservices.py"""
# pylint: disable=C0116
# pylint: disable=W0613

import os
import restate
import services

def test_services():
    names = os.environ.get('SERVICES')
    return services.services_named(names.split(',')) if names else services.all_services()

identity_keys = None
e2e_signing_key_env = os.environ.get('E2E_REQUEST_SIGNING_ENV')
if os.environ.get('E2E_REQUEST_SIGNING_ENV'):
    identity_keys = [os.environ.get('E2E_REQUEST_SIGNING_ENV')]

app = restate.app(services=test_services(), identity_keys=identity_keys)




#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""
This module defines the Endpoint class, which serves as a container for all the services and objects
"""

import typing

from restate.service import Service
from restate.object import VirtualObject
from restate.workflow import Workflow


# disable too few methods in a class
# pylint: disable=R0903


class Endpoint:
    """
    Endpoint service that contains all the services and objects
    """

    services: typing.Dict[str, typing.Union[Service, VirtualObject, Workflow]]
    protocol: typing.Optional[typing.Literal["bidi", "request_response"]]
    identity_keys: typing.List[str]

    def __init__(self):
        """
        Create a new restate endpoint that serves as a container for all the services and objects
        """
        self.services = {}
        # we will let the user to override it later perhaps, but for now let us
        # auto deduce it on discovery.
        # None means that the user did not explicitly set it.
        self.protocol = None

        self.identity_keys = []

    def bind(self, *services: typing.Union[Service, VirtualObject, Workflow]):
        """
        Bind a service to the endpoint

        Args:
            service: The service or virtual object to bind to the endpoint

        Raises:
            ValueError: If a service with the same name already exists in the endpoint

        Returns:
            The updated Endpoint instance
        """
        for service in services:
            if service.name in self.services:
                raise ValueError(f"Service {service.name} already exists")
            if isinstance(service, (Service, VirtualObject, Workflow)):
                self.services[service.name] = service
            else:
                raise ValueError(f"Invalid service type {service}")
        return self

    def streaming_protocol(self):
        """Use bidirectional streaming protocol. Use with servers that support HTTP2"""
        self.protocol = "bidi"
        return self

    def request_response_protocol(self):
        """Use request response style protocol for communication with restate."""
        self.protocol = "request_response"

    def identity_key(self, identity_key: str):
        """Add an identity key to this endpoint."""
        self.identity_keys.append(identity_key)

    def app(self):
        """
        Returns the ASGI application for this endpoint.

        This method is responsible for creating and returning the ASGI application
        that will handle incoming requests for this endpoint.

        Returns:
            The ASGI application for this endpoint.
       """
        # we need to import it here to avoid circular dependencies
        # pylint: disable=C0415
        # pylint: disable=R0401
        from restate.server import asgi_app
        return asgi_app(self)

def app(
    services: typing.Iterable[typing.Union[Service, VirtualObject, Workflow]],
    protocol: typing.Optional[typing.Literal["bidi", "request_response"]] = None,
    identity_keys: typing.Optional[typing.List[str]] = None):
    """A restate ASGI application that hosts the given services."""
    endpoint = Endpoint()
    if protocol == "bidi":
        endpoint.streaming_protocol()
    elif protocol == "request_response":
        endpoint.request_response_protocol()
    for service in services:
        endpoint.bind(service)
    if identity_keys:
        for key in identity_keys:
            endpoint.identity_key(key)
    return endpoint.app()



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE

from typing import Dict, Union
from restate import Service, VirtualObject, Workflow

from .counter import counter_object as s1
from .proxy import proxy as s2
from .awakeable_holder import awakeable_holder as s3
from. block_and_wait_workflow import workflow as s4
from .cancel_test import runner, blocking_service as s5
from .failing import failing as s6
from .kill_test import kill_runner, kill_singleton as s7
from .list_object import list_object as s8
from .map_object import map_object as s9
from .non_determinism import non_deterministic as s10
from .test_utils import test_utils as s11

def list_services(bindings):
    """List all services in this module"""
    return {obj.name : obj for _, obj in bindings.items() if isinstance(obj, (Service, VirtualObject, Workflow))}

def services_named(service_names):
    return [ _all_services[name] for name in service_names ]

def all_services():
    return _all_services.values()

_all_services = list_services(locals())



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
# pylint: disable=R0917
"""
This module defines the Service class for representing a restate service.
"""

from functools import wraps
import inspect
import typing

from restate.serde import Serde, JsonSerde
from .handler import HandlerIO, ServiceTag, make_handler

I = typing.TypeVar('I')
O = typing.TypeVar('O')


# disable too many arguments warning
# pylint: disable=R0913

# disable line too long warning
# pylint: disable=C0301

class VirtualObject:
    """
    Represents a restate virtual object.

    Args:
        name (str): The name of the object.
    """

    def __init__(self, name):
        self.service_tag = ServiceTag("object", name)
        self.handlers = {}

    @property
    def name(self):
        """
        Returns the name of the object.
        """
        return self.service_tag.name

    def handler(self,
                name: typing.Optional[str] = None,
                kind: typing.Optional[typing.Literal["exclusive", "shared"]] = "exclusive",
                accept: str = "application/json",
                content_type: str = "application/json",
                input_serde: Serde[I] = JsonSerde[I](), # type: ignore
                output_serde: Serde[O] = JsonSerde[O]()) -> typing.Callable: # type: ignore
        """
        Decorator for defining a handler function.

        Args:
            name: The name of the handler. 
            accept: The accept type of the request. Default "application/json".
            content_type: The content type of the request. Default "application/json".
            serializer: The serializer function to convert the response object to bytes. 
            deserializer: The deserializer function to convert the request bytes to an object.

        Returns:
            Callable: The decorated function.

        Raises:
            ValueError: If the handler name is not provided.

        Example:
            @service.handler()
            def my_handler_func(ctx, request):
                # handler logic
                pass
        """
        handler_io = HandlerIO[I,O](accept, content_type, input_serde, output_serde)
        def wrapper(fn):

            @wraps(fn)
            def wrapped(*args, **kwargs):
                return fn(*args, **kwargs)

            signature = inspect.signature(fn, eval_str=True)
            handler = make_handler(self.service_tag, handler_io, name, kind, wrapped, signature)
            self.handlers[handler.name] = handler
            return wrapped

        return wrapper
